/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011-2015 FlowKit Sarl
 * Route d'Oron 2
 * 1010 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef ADVECTION_DIFFUSION_PROCESSOR_2D_H
#define ADVECTION_DIFFUSION_PROCESSOR_2D_H

#include "core/globalDefs.h"
#include "core/dynamics.h"
#include "atomicBlock/dataProcessingFunctional2D.h"


namespace plb {

// This data processor uses asymmetric finite differences to compute a gradient.
template<typename T, template<typename U> class Descriptor, int xNormal, int yNormal>
class AdvectionDiffusionCornerBoundaryProcessor2D : public BoxProcessingFunctional2D_L<T,Descriptor>
{
public:
    virtual void process(Box2D domain, BlockLattice2D<T,Descriptor>& lattice);
    
    virtual AdvectionDiffusionCornerBoundaryProcessor2D<T,Descriptor,xNormal,yNormal>* clone() const {
        return new AdvectionDiffusionCornerBoundaryProcessor2D<T,Descriptor,xNormal,yNormal>(*this);
    }
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const {
        modified[0] = modif::staticVariables;   // lattice
    }
};

// This data processor uses asymmetric finite differences to compute a gradient.
template<typename T, template<typename U> class Descriptor, int xNormal, int yNormal>
class CompleteAdvectionDiffusionCornerBoundaryProcessor2D : public BoxProcessingFunctional2D_L<T,Descriptor>
{
public:
    virtual void process(Box2D domain, BlockLattice2D<T,Descriptor>& lattice);
    
    virtual CompleteAdvectionDiffusionCornerBoundaryProcessor2D<T,Descriptor,xNormal,yNormal>* clone() const {
        return new CompleteAdvectionDiffusionCornerBoundaryProcessor2D<T,Descriptor,xNormal,yNormal>(*this);
    }
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const {
        modified[0] = modif::staticVariables;   // lattice
    }
};


    
} // namespace plb

#endif  // ADVECTION_DIFFUSION_PROCESSOR_2D_H

